"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: Apache-2.0
"""

# Export public facing objects
# flake8: noqa

from .streammanagerclient import StreamManagerClient, SDK_VERSION
from .exceptions import *
from .util import Util
from .data import (
    ReadMessagesOptions,
    MessageStreamDefinition,
    ExportDefinition,
    StrategyOnFull,
    Persistence,
    HTTPConfig,
    IoTAnalyticsConfig,
    KinesisConfig,
    ExportFormat,
    # Status related
    # Config
    StatusConfig,
    # Data
    StatusContext,
    StatusLevel,
    EventType,
    Status,
    StatusMessage,
    # S3 Tasks related:
    # Config
    S3ExportTaskExecutorConfig,
    # Data
    S3ExportTaskDefinition,
    # Iot SiteWise related:
    # Config
    IoTSiteWiseConfig,
    # Data
    Variant,
    Quality,
    TimeInNanos,
    AssetPropertyValue,
    PutAssetPropertyValueEntry,
)



#
# Copyright 2010-2022 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#

from .client import client
# .Lambda imports greengrass_common, which only applies within Greengrass Core.
# Try-except as below to make sure the SDK is able to be imported outside of Greengrass Core
try:
    from .Lambda import StreamingBody
except:
    pass

__version__ = '1.6.1'
INTERFACE_VERSION = '1.5'



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: Apache-2.0
"""

import asyncio
import json
import re
import uuid
from typing import Sequence

from .data import ResponseStatusCode
from .exceptions import (
    InvalidRequestException,
    MessageStoreReadErrorException,
    NotEnoughMessagesException,
    RequestPayloadTooLargeException,
    ResourceNotFoundException,
    ResponsePayloadTooLargeException,
    ServerOutOfMemoryException,
    ServerTimeoutException,
    StreamManagerException,
    UnauthorizedException,
    UnknownFailureException,
    UnknownOperationException,
    UpdateFailedException,
    UpdateNotAllowedException,
)


class UtilInternal:
    __ENDIAN = "big"
    _MAX_PACKET_SIZE = 1 << 30

    @staticmethod
    def sync(coro, loop: asyncio.AbstractEventLoop):
        if asyncio.iscoroutine(coro):
            # Run async function in the loop and return the value or raise the exception
            return asyncio.run_coroutine_threadsafe(coro, loop=loop).result()

        return coro

    """
    Delete keys with the value ``None`` in a dictionary, recursively.
    This alters the input so you may wish to ``copy`` the dict first.
    """

    @staticmethod
    def del_empty_arrays(d):

        for key, value in list(d.items()):
            if isinstance(value, list) and len(value) == 0:
                del d[key]
            elif isinstance(value, dict):
                UtilInternal.del_empty_arrays(value)
        return d

    @staticmethod
    def serialize_to_json_with_empty_array_as_null(data):
        s = json.dumps(UtilInternal.del_empty_arrays(data.as_dict()))

        return s.encode()

    @staticmethod
    def int_to_bytes(i, length=4):
        return int.to_bytes(i, length=length, byteorder=UtilInternal.__ENDIAN, signed=True)

    @staticmethod
    def int_from_bytes(b):
        return int.from_bytes(b, byteorder=UtilInternal.__ENDIAN, signed=True)

    @staticmethod
    def encode_frame(frame) -> Sequence[bytes]:
        if len(frame.payload) + 1 > UtilInternal._MAX_PACKET_SIZE:
            raise RequestPayloadTooLargeException()
        return [
            bytes(
                [
                    *UtilInternal.int_to_bytes(len(frame.payload) + 1),
                    *UtilInternal.int_to_bytes(frame.operation.value, length=1),
                ]
            ),
            frame.payload,
        ]

    @staticmethod
    def get_request_id():
        return str(uuid.uuid4())

    @staticmethod
    def is_invalid(o):
        if not hasattr(o, "_validations_map"):
            return False
        if not hasattr(o, "_types_map"):
            return False
        for prop_name, validations in o._validations_map.items():
            if not hasattr(o, prop_name):
                return "Object is malformed, missing property: {}".format(prop_name)
            # Validate all properties on lists
            if type(getattr(o, prop_name)) == list:
                for i, v in enumerate(getattr(o, prop_name)):
                    result = UtilInternal.is_invalid(v)
                    if result:
                        return "Property {}[{}] is invalid because {}".format(prop_name, i, result)

            # Recurse down to check validity of objects within objects
            result = UtilInternal.is_invalid(getattr(o, prop_name))
            if result:
                return "Property {} is invalid because {}".format(prop_name, result)

            # Validate the property
            if "required" in validations and validations["required"] and getattr(o, prop_name) is None:
                return "Property {} is required, but was None".format(prop_name)
            if (
                "minLength" in validations
                and getattr(o, prop_name) is not None
                and len(getattr(o, prop_name)) < validations["minLength"]
            ):
                return "Property {} must have a minimum length of {}, but found length of {}".format(
                    prop_name, validations["minLength"], len(getattr(o, prop_name))
                )
            if (
                "maxLength" in validations
                and getattr(o, prop_name) is not None
                and len(getattr(o, prop_name)) > validations["maxLength"]
            ):
                return "Property {} must have a maximum length of {}, but found length of {}".format(
                    prop_name, validations["maxLength"], len(getattr(o, prop_name))
                )
            if (
                "minItems" in validations
                and getattr(o, prop_name) is not None
                and len(getattr(o, prop_name)) < validations["minItems"]
            ):
                return "Property {} must have at least {} items, but found {}".format(
                    prop_name, validations["minItems"], len(getattr(o, prop_name))
                )
            if (
                "maxItems" in validations
                and getattr(o, prop_name) is not None
                and len(getattr(o, prop_name)) > validations["maxItems"]
            ):
                return "Property {} must have at most {} items, but found {}".format(
                    prop_name, validations["maxItems"], len(getattr(o, prop_name))
                )
            if (
                "maximum" in validations
                and getattr(o, prop_name) is not None
                and getattr(o, prop_name) > validations["maximum"]
            ):
                return "Property {} must be at most {}".format(prop_name, validations["maximum"])
            if (
                "minimum" in validations
                and getattr(o, prop_name) is not None
                and getattr(o, prop_name) < validations["minimum"]
            ):
                return "Property {} must be at least {}".format(prop_name, validations["minimum"])
            if (
                "pattern" in validations
                and getattr(o, prop_name) is not None
                and re.fullmatch(validations["pattern"], getattr(o, prop_name)) is None
            ):
                return "Property {} must match regex {}".format(prop_name, validations["pattern"])

        for prop_name, types in o._types_map.items():
            # Validate all properties with their respective types
            if "type" in types and getattr(o, prop_name) is not None:
                result = isinstance(getattr(o, prop_name), types["type"])
                if not result:
                    return "Property {} is invalid because it must be of type {}".format(
                        prop_name, types["type"].__name__
                    )
                if types["type"] == list and "subtype" in types:
                    for i, v in enumerate(getattr(o, prop_name)):
                        result = isinstance(v, types["subtype"])
                        if not result:
                            return "Property {}[{}] is invalid because it must be of type {}".format(
                                prop_name, i, types["subtype"].__name__
                            )

        return False

    @staticmethod
    def raise_on_error_response(response):
        if response.status == ResponseStatusCode.Success:
            return
        elif response.status == ResponseStatusCode.InvalidRequest:
            raise InvalidRequestException(response.error_message, response.status, response.request_id)
        elif response.status == ResponseStatusCode.RequestPayloadTooLarge:
            raise RequestPayloadTooLargeException(response.error_message, response.status, response.request_id)
        elif response.status == ResponseStatusCode.ResourceNotFound:
            raise ResourceNotFoundException(response.error_message, response.status, response.request_id)
        elif response.status == ResponseStatusCode.ResponsePayloadTooLarge:
            raise ResponsePayloadTooLargeException(response.error_message, response.status, response.request_id)
        elif response.status == ResponseStatusCode.ServerTimeout:
            raise ServerTimeoutException(response.error_message, response.status, response.request_id)
        elif response.status == ResponseStatusCode.Unauthorized:
            raise UnauthorizedException(response.error_message, response.status, response.request_id)
        elif response.status == ResponseStatusCode.UnknownFailure:
            raise UnknownFailureException(response.error_message, response.status, response.request_id)
        elif response.status == ResponseStatusCode.NotEnoughMessages:
            raise NotEnoughMessagesException(response.error_message, response.status, response.request_id)
        elif response.status == ResponseStatusCode.MessageStoreReadError:
            raise MessageStoreReadErrorException(response.error_message, response.status, response.request_id)
        elif response.status == ResponseStatusCode.OutOfMemoryError:
            raise ServerOutOfMemoryException(response.error_message, response.status, response.request_id)
        elif response.status == ResponseStatusCode.UpdateFailed:
            raise UpdateFailedException(response.error_message, response.status, response.request_id)
        elif response.status == ResponseStatusCode.UpdateNotAllowed:
            raise UpdateNotAllowedException(response.error_message, response.status, response.request_id)
        elif response.status == ResponseStatusCode.UnknownOperation:
            raise UnknownOperationException(response.error_message, response.status, response.request_id)
        else:
            raise StreamManagerException(
                "Client is not able to understand this server response status code", "Unrecognized", response.request_id
            )



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#


def client(client_type, *args, **kwargs):
    if client_type == 'lambda':
        from .Lambda import Client
    elif client_type == 'iot-data':
        from .IoTDataPlane import Client
    elif client_type == 'secretsmanager':
        from .SecretsManager import Client
    elif client_type == 'streammanager':
        from .stream_manager import StreamManagerClient as Client
    else:
        raise Exception('Client type {} is not recognized.'.format(repr(client_type)))

    return Client(*args, **kwargs)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: Apache-2.0
"""

import asyncio
import logging
import os
from threading import Thread
from typing import List, Optional

import cbor2

from .data import (
    AppendMessageRequest,
    AppendMessageResponse,
    ConnectRequest,
    ConnectResponse,
    CreateMessageStreamRequest,
    CreateMessageStreamResponse,
    DeleteMessageStreamRequest,
    DeleteMessageStreamResponse,
    DescribeMessageStreamRequest,
    DescribeMessageStreamResponse,
    ListStreamsRequest,
    ListStreamsResponse,
    Message,
    MessageFrame,
    MessageStreamDefinition,
    MessageStreamInfo,
    Operation,
    ReadMessagesOptions,
    ReadMessagesRequest,
    ReadMessagesResponse,
    ResponseStatusCode,
    UnknownOperationError,
    UpdateMessageStreamRequest,
    UpdateMessageStreamResponse,
    VersionInfo,
)
from .exceptions import ClientException, ConnectFailedException, StreamManagerException, ValidationException
from .utilinternal import UtilInternal

# Version of the Python SDK.
# NOTE: This version is independent of the StreamManager PROTOCOL_VERSION, which versions the data format
#  over the wire. When bumping the PROTOCOL_VERSION, consider adding the old version to
#  __OLD_SUPPORTED_PROTOCOL_VERSIONS list (if you intend to support it). Nothing else is needed to bump
#  this SDK_VERSION.
SDK_VERSION = "1.1.1"


class StreamManagerClient:
    """
    Creates a client for the Greengrass StreamManager. All parameters are optional.

    :param host: The host which StreamManager server is running on. Default is localhost.
    :param port: The port which StreamManager server is running on. Default is found in environment variables.
    :param connect_timeout: The timeout in seconds for connecting to the server. Default is 3 seconds.
    :param request_timeout: The timeout in seconds for all operations. Default is 60 seconds.
    :param logger: A logger to use for client logging. Default is Python's builtin logger.

    :raises: :exc:`~.exceptions.StreamManagerException` and subtypes if authenticating to the server fails.
    :raises: :exc:`asyncio.TimeoutError` if the request times out.
    :raises: :exc:`ConnectionError` if the client is unable to connect to the server.
    """

    # List of supported protocol protocol.
    # These are meant to be used for graceful degradation if the server does not support the current SDK version.
    __OLD_SUPPORTED_PROTOCOL_VERSIONS = ["1.0.0"]

    __CONNECT_VERSION = 1

    def __init__(
        self,
        host="127.0.0.1",
        port=None,
        connect_timeout=3,
        request_timeout=60,
        logger=logging.getLogger("StreamManagerClient"),
    ):
        self.host = host
        if port is None:
            port = int(os.getenv("STREAM_MANAGER_SERVER_PORT", 8088))
        self.port = port
        self.__requests = {}
        self.connect_timeout = connect_timeout
        self.request_timeout = request_timeout
        self.logger = logger
        self.auth_token = os.getenv("AWS_CONTAINER_AUTHORIZATION_TOKEN")

        # Python Logging doesn't have a TRACE level
        # so we will add our own at level 5. (Debug is level 10)
        if logger.level <= 5:
            logging.addLevelName(5, "TRACE")

        self.__loop = asyncio.new_event_loop()
        self.__closed = False
        self.__reader = None
        self.__writer = None

        # Defines a function to be run in a separate thread to run the event loop
        # this enables our synchronous interface without locks
        def run_event_loop(loop: asyncio.AbstractEventLoop):
            try:
                loop.run_forever()
            finally:
                loop.close()

        # Making the thread a daemon will kill the thread once the main thread closes
        self.__event_loop_thread = Thread(target=run_event_loop, args=(self.__loop,), daemon=True)
        self.__event_loop_thread.start()

        self.connected = False
        UtilInternal.sync(self.__connect(), loop=self.__loop)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    async def _close(self):
        if self.__writer is not None:
            self.__closed = True
            self.connected = False
            self.__reader = None
            # Drain any existing data waiting to be sent
            await self.__writer.drain()
            self.__writer.close()
            try:
                # Only added in Python 3.7, so try to call it, but otherwise just skip it
                await self.__writer.wait_closed()
            except AttributeError:
                pass
            self.__writer = None

    def __check_closed(self):
        if self.__closed:
            raise StreamManagerException("Client is closed. Create a new client first.")

    async def __connect(self):
        self.__check_closed()
        if self.connected:
            return
        try:
            self.logger.debug("Opening connection to %s:%d", self.host, self.port)
            future = asyncio.open_connection(self.host, self.port)
            self.__reader, self.__writer = await asyncio.wait_for(
                future, timeout=self.connect_timeout
            )

            await asyncio.wait_for(self.__connect_request_response(), timeout=self.request_timeout)

            self.logger.debug("Socket connected successfully. Starting read loop.")
            self.connected = True
            self.__loop.create_task(self.__read_loop())
        except ConnectionError as e:
            self.logger.error("Connection error while connecting to server: %s", e)
            raise

    def __log_trace(self, *args, **kwargs):
        self.logger.log(5, *args, **kwargs)

    async def __read_message_frame(self):
        length_bytes = await self.__reader.read(n=4)
        if len(length_bytes) == 0:
            raise asyncio.IncompleteReadError(length_bytes, 4)

        length = UtilInternal.int_from_bytes(length_bytes)
        operation = UtilInternal.int_from_bytes(await self.__reader.read(n=1))

        # Read from the socket until we have read the full packet
        payload = bytearray()
        read_bytes = 1
        while read_bytes < length:
            next_payload = await self.__reader.read(n=length - read_bytes)
            if len(next_payload) == 0:
                raise asyncio.IncompleteReadError(next_payload, length - read_bytes)
            payload.extend(next_payload)
            read_bytes += len(next_payload)

        try:
            op = Operation.from_dict(operation)
        except ValueError:
            self.logger.error("Found unknown operation %d", operation)
            op = Operation.Unknown

        return MessageFrame(operation=op, payload=bytes(payload))

    async def __read_loop(self):
        # Continually try to read packets from the socket
        while not self.__closed:
            try:
                try:
                    self.__log_trace("Starting long poll read")
                    response = await self.__read_message_frame()
                    self.__log_trace("Got message frame from server: %s", response)
                except asyncio.IncompleteReadError:
                    if self.__closed:
                        return
                    self.logger.error("Unable to read from socket, likely socket is closed or server died")
                    self.connected = False
                    try:
                        await self.__connect()
                    except ConnectionError:
                        # Already logged in __connect, so just ignore it here
                        pass
                    except ConnectFailedException:
                        # Already logged in __connect_request_response, so just ignore it here
                        pass
                    return

                payload = cbor2.loads(response.payload)
                await self.__handle_read_response(payload, response)
            except Exception:
                self.logger.exception("Unhandled exception occurred")
                return

    async def __handle_read_response(self, payload, response):
        if response.operation == Operation.ReadMessagesResponse:
            response = ReadMessagesResponse.from_dict(payload)
            self.logger.debug("Received ReadMessagesResponse from server")
            await self.__requests[response.request_id].put(response)
        elif response.operation == Operation.CreateMessageStreamResponse:
            response = CreateMessageStreamResponse.from_dict(payload)
            self.logger.debug("Received CreateMessageStreamResponse from server: %s", response)
            await self.__requests[response.request_id].put(response)
        elif response.operation == Operation.DeleteMessageStreamResponse:
            response = DeleteMessageStreamResponse.from_dict(payload)
            self.logger.debug("Received DeleteMessageStreamResponse from server: %s", response)
            await self.__requests[response.request_id].put(response)
        elif response.operation == Operation.UpdateMessageStreamResponse:
            response = UpdateMessageStreamResponse.from_dict(payload)
            self.logger.debug("Received UpdateMessageStreamResponse from server: %s", response)
            await self.__requests[response.request_id].put(response)
        elif response.operation == Operation.AppendMessageResponse:
            response = AppendMessageResponse.from_dict(payload)
            self.logger.debug("Received AppendMessageResponse from server: %s", response)
            await self.__requests[response.request_id].put(response)
        elif response.operation == Operation.ListStreamsResponse:
            response = ListStreamsResponse.from_dict(payload)
            self.logger.debug("Received ListStreamsResponse from server: %s", response)
            await self.__requests[response.request_id].put(response)
        elif response.operation == Operation.DescribeMessageStreamResponse:
            response = DescribeMessageStreamResponse.from_dict(payload)
            self.logger.debug("Received DescribeMessageStreamResponse from server: %s", response)
            await self.__requests[response.request_id].put(response)
        elif response.operation == Operation.UnknownOperationError:
            self.logger.error(
                "Received response with unsupported operation from server: %s. "
                "You should update your server version",
                response.operation,
            )
            response = UnknownOperationError.from_dict(payload)
            await self.__requests[response.request_id].put(response)
        elif response.operation == Operation.Unknown:
            self.logger.error("Received response with unknown operation from server: %s", response)
            try:
                request_id = cbor2.loads(response.payload)["requestId"]
                await self.__requests[request_id].put(response)
            except Exception:
                # We tried our best to figure out the request id, but it failed.
                # We already logged the unknown operation, so there's nothing
                # else we can do at this point
                pass
        else:
            self.logger.error("Received data with unhandled operation %s.", response.operation)

    async def __connect_request_response(self):
        data = ConnectRequest()
        data.request_id = UtilInternal.get_request_id()
        data.sdk_version = SDK_VERSION
        data.other_supported_protocol_versions = self.__OLD_SUPPORTED_PROTOCOL_VERSIONS
        data.protocol_version = VersionInfo.PROTOCOL_VERSION.value
        if self.auth_token is not None:
            data.auth_token = self.auth_token

        # Write the connect version
        self.__writer.write(UtilInternal.int_to_bytes(self.__CONNECT_VERSION, 1))
        # Write request to socket
        frame = MessageFrame(operation=Operation.Connect, payload=cbor2.dumps(data.as_dict()))
        for b in UtilInternal.encode_frame(frame):
            self.__writer.write(b)
        await self.__writer.drain()

        # Read connect version
        connect_response_version_byte = await self.__reader.read(n=1)
        if len(connect_response_version_byte) == 0:
            raise asyncio.IncompleteReadError(connect_response_version_byte, 1)

        connect_response_version = UtilInternal.int_from_bytes(connect_response_version_byte)
        if connect_response_version != self.__CONNECT_VERSION:
            self.logger.error("Unexpected response from the server, Connect version: %s.", connect_response_version)
            raise ConnectFailedException("Failed to establish connection with the server")

        # Read connect response
        response = await self.__read_message_frame()  # type: MessageFrame

        if response.operation == Operation.ConnectResponse:
            payload = cbor2.loads(response.payload)
            response = ConnectResponse.from_dict(payload)  # type: ConnectResponse
            self.logger.debug("Received ConnectResponse from server: %s", response)
        else:
            self.logger.error("Received data with unexpected operation %s.", response.operation)
            raise ConnectFailedException("Failed to establish connection with the server")

        if response.status != ResponseStatusCode.Success:
            self.logger.error("Received ConnectResponse with unexpected status %s.", response.status)
            raise ConnectFailedException("Failed to establish connection with the server")

        if data.protocol_version != response.protocol_version:
            self.logger.warn(
                "SDK with version %s using Protocol version %s is not fully compatible with Server with version %s. "
                "Client has connected in a compatibility mode using protocol version %s. "
                "Some features will not work as expected",
                SDK_VERSION,
                data.protocol_version,
                response.server_version,
                response.protocol_version,
            )

    async def __send_and_receive(self, operation, data):
        async def inner(operation, data):
            if data.request_id is None:
                data.request_id = UtilInternal.get_request_id()

            validation = UtilInternal.is_invalid(data)
            if validation:
                raise ValidationException(validation)

            # If we're not connected, immediately try to reconnect
            if not self.connected:
                await self.__connect()

            self.__requests[data.request_id] = asyncio.Queue(1)

            # Write request to socket
            frame = MessageFrame(operation=operation, payload=cbor2.dumps(data.as_dict()))
            for b in UtilInternal.encode_frame(frame):
                self.__writer.write(b)
            await self.__writer.drain()

            # Wait for reader to come back with the response
            result = await self.__requests[data.request_id].get()
            # Drop async queue from request map
            del self.__requests[data.request_id]
            if isinstance(result, MessageFrame) and result.operation == Operation.Unknown:
                raise ClientException("Received response with unknown operation from server")
            return result

        # Perform the actual work as async so that we can put a timeout on the whole operation
        try:
            return await asyncio.wait_for(inner(operation, data), timeout=self.request_timeout)
        except asyncio.TimeoutError:
            # Drop async queue from request map
            del self.__requests[data.request_id]
            raise

    def __validate_read_message_options(self, options: Optional[ReadMessagesOptions]):
        if options is not None:
            if not isinstance(options, ReadMessagesOptions):
                raise ValidationException("options argument to read_messages must be a ReadMessageOptions object")
            validation = UtilInternal.is_invalid(options)
            if validation:
                raise ValidationException(validation)
            if (
                options.min_message_count is not None
                and options.max_message_count is not None
                and options.min_message_count > options.max_message_count
            ):
                raise ValidationException("min_message_count must be less than or equal to max_message_count")
            if options.read_timeout_millis is not None and options.read_timeout_millis > self.request_timeout * 1000:
                raise ValidationException(
                    "read_timeout_millis must be less than or equal to the client's request_timeout"
                )

    async def _append_message(self, stream_name: str, data: bytes) -> int:
        append_message_request = AppendMessageRequest(name=stream_name, payload=data)
        append_message_response = await self.__send_and_receive(
            Operation.AppendMessage, data=append_message_request
        )  # type: AppendMessageResponse

        UtilInternal.raise_on_error_response(append_message_response)
        return append_message_response.sequence_number

    async def _create_message_stream(self, definition: MessageStreamDefinition) -> None:
        if not isinstance(definition, MessageStreamDefinition):
            raise ValidationException("definition argument to create_stream must be a MessageStreamDefinition object")
        create_stream_request = CreateMessageStreamRequest(definition=definition)
        create_stream_response = await self.__send_and_receive(
            Operation.CreateMessageStream, data=create_stream_request
        )  # type: CreateMessageStreamResponse

        UtilInternal.raise_on_error_response(create_stream_response)

    async def _delete_message_stream(self, stream_name: str) -> None:
        delete_stream_request = DeleteMessageStreamRequest(name=stream_name)
        delete_stream_response = await self.__send_and_receive(
            Operation.DeleteMessageStream, data=delete_stream_request
        )  # type: DeleteMessageStreamResponse

        UtilInternal.raise_on_error_response(delete_stream_response)

    async def _update_message_stream(self, definition: MessageStreamDefinition) -> None:
        if not isinstance(definition, MessageStreamDefinition):
            raise ValidationException(
                "definition argument to update_message_stream must be a MessageStreamDefinition object"
            )
        update_stream_request = UpdateMessageStreamRequest(definition=definition)
        update_stream_response = await self.__send_and_receive(
            Operation.UpdateMessageStream, data=update_stream_request
        )  # type: UpdateMessageStreamResponse

        UtilInternal.raise_on_error_response(update_stream_response)

    async def _read_messages(self, stream_name: str, options: ReadMessagesOptions = None) -> List[Message]:
        self.__validate_read_message_options(options)
        read_messages_request = ReadMessagesRequest(stream_name=stream_name, read_messages_options=options)
        read_messages_response = await self.__send_and_receive(
            Operation.ReadMessages, data=read_messages_request
        )  # type: ReadMessagesResponse

        UtilInternal.raise_on_error_response(read_messages_response)
        return read_messages_response.messages

    async def _list_streams(self) -> List[str]:
        list_streams_response = await self.__send_and_receive(
            Operation.ListStreams, data=ListStreamsRequest()
        )  # type: ListStreamsResponse

        UtilInternal.raise_on_error_response(list_streams_response)
        return list_streams_response.streams

    async def _describe_message_stream(self, stream_name: str) -> MessageStreamInfo:
        describe_message_stream_response = await self.__send_and_receive(
            Operation.DescribeMessageStream, data=DescribeMessageStreamRequest(name=stream_name)
        )  # type: DescribeMessageStreamResponse
        UtilInternal.raise_on_error_response(describe_message_stream_response)

        return describe_message_stream_response.message_stream_info

    ####################
    #    PUBLIC API    #
    ####################
    def read_messages(self, stream_name: str, options: Optional[ReadMessagesOptions] = None) -> List[Message]:
        """
        Read message(s) from a chosen stream with options. If no options are specified it will try to read
        1 message from the stream.

        :param stream_name: The name of the stream to read from.
        :param options: (Optional) Options used when reading from the stream of type :class:`.data.ReadMessagesOptions`.
            Defaults are:

            * desired_start_sequence_number: 0,
            * min_message_count: 1,
            * max_message_count: 1,
            * read_timeout_millis: 0 ``# Where 0 here represents that the server will immediately return the messages``
                                     ``# or an exception if there were not enough messages available.``
            If desired_start_sequence_number is specified in the options and is less
            than the current beginning of the stream, returned messages will start
            at the beginning of the stream and not necessarily the desired_start_sequence_number.
        :return: List of at least 1 message.
        :raises: :exc:`~.exceptions.StreamManagerException` and subtypes based on the precise error.
        :raises: :exc:`asyncio.TimeoutError` if the request times out.
        :raises: :exc:`ConnectionError` if the client is unable to reconnect to the server.
        """
        self.__check_closed()
        return UtilInternal.sync(self._read_messages(stream_name, options), loop=self.__loop)

    def append_message(self, stream_name: str, data: bytes) -> int:
        """
        Append a message into the specified message stream. Returns the sequence number of the message
        if it was successfully appended.

        :param stream_name: The name of the stream to append to.
        :param data: Bytes type data.
        :return: Sequence number that the message was assigned if it was appended.
        :raises: :exc:`~.exceptions.StreamManagerException` and subtypes based on the precise error.
        :raises: :exc:`asyncio.TimeoutError` if the request times out.
        :raises: :exc:`ConnectionError` if the client is unable to reconnect to the server.
        """
        self.__check_closed()
        return UtilInternal.sync(self._append_message(stream_name, data), loop=self.__loop)

    def create_message_stream(self, definition: MessageStreamDefinition) -> None:
        """
        Create a message stream with a given definition.

        :param definition: :class:`~.data.MessageStreamDefinition` definition object.
        :return: Nothing is returned if the request succeeds.
        :raises: :exc:`~.exceptions.StreamManagerException` and subtypes based on the precise error.
        :raises: :exc:`asyncio.TimeoutError` if the request times out.
        :raises: :exc:`ConnectionError` if the client is unable to reconnect to the server.
        """
        self.__check_closed()
        return UtilInternal.sync(self._create_message_stream(definition), loop=self.__loop)

    def delete_message_stream(self, stream_name: str) -> None:
        """
        Deletes a message stream based on its name. Nothing is returned if the request succeeds,
        a subtype of :exc:`~.exceptions.StreamManagerException` will be raised if an error occurs.

        :param stream_name: The name of the stream to be deleted.
        :return: Nothing is returned if the request succeeds.
        :raises: :exc:`~.exceptions.StreamManagerException` and subtypes based on the precise error.
        :raises: :exc:`asyncio.TimeoutError` if the request times out.
        :raises: :exc:`ConnectionError` if the client is unable to reconnect to the server.
        """
        self.__check_closed()
        return UtilInternal.sync(self._delete_message_stream(stream_name), loop=self.__loop)

    def update_message_stream(self, definition: MessageStreamDefinition) -> None:
        """
        Updates a message stream based on a given definition.
        Minimum version requirements: StreamManager server version 1.1 (or AWS IoT Greengrass Core 1.11.0)

        :param definition: class:`~.data.MessageStreamDefinition` definition object.
        :return: Nothing is returned if the request succeeds.
        :raises: :exc:`~.exceptions.StreamManagerException` and subtypes based on the precise error.
        :raises: :exc:`asyncio.TimeoutError` if the request times out.
        :raises: :exc:`ConnectionError` if the client is unable to reconnect to the server.
        """
        self.__check_closed()
        return UtilInternal.sync(self._update_message_stream(definition), loop=self.__loop)

    def list_streams(self) -> List[str]:
        """
        List the streams in StreamManager. Returns a list of their names.

        :return: List of stream names.
        :raises: :exc:`~.exceptions.StreamManagerException` and subtypes based on the precise error.
        :raises: :exc:`asyncio.TimeoutError` if the request times out.
        :raises: :exc:`ConnectionError` if the client is unable to reconnect to the server.
        """
        self.__check_closed()
        return UtilInternal.sync(self._list_streams(), loop=self.__loop)

    def describe_message_stream(self, stream_name: str) -> MessageStreamInfo:
        """
        Describe a message stream to get metadata including the stream's definition,
        size, and exporter statuses.

        :param stream_name: The name of the stream to describe.
        :return: :class:`~.data.MessageStreamInfo` type containing the stream information.
        :raises: :exc:`~.exceptions.StreamManagerException` and subtypes based on the precise error.
        :raises: :exc:`asyncio.TimeoutError` if the request times out.
        :raises: :exc:`ConnectionError` if the client is unable to reconnect to the server.
        """
        self.__check_closed()
        return UtilInternal.sync(self._describe_message_stream(stream_name), loop=self.__loop)

    def close(self):
        """
        Call to shutdown the client and close all existing connections. Once a client is closed it cannot be reused.

        :raises: :exc:`~.exceptions.StreamManagerException` and subtypes based on the precise error.
        """
        if not self.__closed:
            UtilInternal.sync(self._close(), loop=self.__loop)
        if not self.__loop.is_closed():
            self.__loop.call_soon_threadsafe(self.__loop.stop)



#!/usr/bin/env python

"""
distutils/setuptools install script.
"""
import os
import re

from setuptools import setup, find_packages


ROOT = os.path.dirname(__file__)
VERSION_RE = re.compile(r'''__version__ = ['"]([0-9.]+)['"]''')


requires = [
    "cbor2~=5.4.2"
]


def get_version():
    init = open(os.path.join(ROOT,'greengrasssdk','__init__.py')).read()
    version = VERSION_RE.search(init).group(1)
    if os.getenv("PYPI_TEST", False) == "True":
        import time
        return version + "." + str(int(time.time()))
    return version


setup(
    name='greengrasssdk',
    version=get_version(),
    description='The AWS IoT Greengrass SDK for Python',
    long_description=open('README.rst').read(),
    author='Amazon Web Services',
    url='',
    scripts=[],
    packages=find_packages(),
    package_data={
        'greengrasssdk': [
        ]
    },
    include_package_data=True,
    install_requires=requires,
    license="Apache License 2.0",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.7',
    ],
)



#
# Copyright 2010-2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#

import logging
import re

from io import BytesIO

from greengrass_common.function_arn_fields import FunctionArnFields
from greengrass_ipc_python_sdk.ipc_client import IPCClient, IPCException
from greengrasssdk.utils.testing import mock

# Log messages in the SDK are part of customer's log because they're helpful for debugging
# customer's lambdas. Since we configured the root logger to log to customer's log and set the
# propagate flag of this logger to True. The log messages submitted from this logger will be
# sent to the customer's local Cloudwatch handler.
customer_logger = logging.getLogger(__name__)
customer_logger.propagate = True

valid_base64_regex = '^([A-Za-z0-9+/]{4})*([A-Za-z0-9+/]{4}|[A-Za-z0-9+/]{3}=|[A-Za-z0-9+/]{2}==)$'


class InvocationException(Exception):
    pass


class Client:
    def __init__(self, endpoint='localhost', port=None):
        """
        :param endpoint: Endpoint used to connect to IPC.
        :type endpoint: str

        :param port: Deprecated. Will not be used.
        :type port: None
        """
        self.ipc = IPCClient(endpoint=endpoint)

    def invoke(self, **kwargs):
        r"""
        Invokes Lambda function of the given name.

        :Keyword Arguments:
            * *ClientContext* (``bytes``) --
              Optional Base64-encoded data about the invoking client to pass to the Lambda function
            * *FunctionName* (``string``) --
              [REQUIRED]
              The Amazon Resource Name (ARN) of the Lambda function to invoke. Name formats:

              * Qualified ARN - The function ARN with the version suffix. e.g. arn:aws:lambda:aws-region:acct-id:function:helloworld:1
              * Unqualified ARN - The function ARN without the version suffix. e.g. arn:aws:lambda:aws-region:acct-id:function:helloworld
            * *InvocationType* (``string``) --
              Choose from the following options.

              * ``RequestResponse`` (default) - Invoke the Lambda synchronously. Block until the function returns a response or times out.
              * ``Event`` - Invoke the Lambda asynchronously. The response only includes empty payload.
            * *Payload* (``bytes``) --
              Optional input for the Lambda function to invoke.
            * *Qualifier* (``string``) --
              Optional parameter to specify a Lambda function version if it was not included in the FunctionName field.
              If you specify a function version, the API uses the qualified function ARN to invoke a specific Lambda function.
        :returns: (``dict``) --
            * *FunctionError* (``string``) --
              If present, indicates that an error occurred while executing the Lambda function. If an error occurred,
              this field will have one of two values, ``Handled`` or ``Unhandled``. ``Handled`` errors are errors that are reported by the function
              while the ``Unhandled`` errors are those detected and reported by Greengrass Core.
              ``Unhandled`` errors include out of memory errors and function timeouts. Error details are provided in the Payload.
            * *Payload* (``bytes or StreamingBody object``) --
              It is the result returned by the Lambda function. This is present only if the invocation type is ``RequestResponse``.
              In the event of a function error this field contains a message describing the error.
        """

        # FunctionName is a required parameter
        if 'FunctionName' not in kwargs:
            raise ValueError(
                '"FunctionName" argument of Lambda.Client.invoke is a required argument but was not provided.'
            )

        arn_fields = FunctionArnFields(kwargs['FunctionName'])
        arn_qualifier = arn_fields.qualifier

        # A Function qualifier can be provided as part of the ARN in FunctionName, or it can be provided here. The
        # behavior of the cloud is to throw an exception if both are specified but not equal
        extraneous_qualifier = kwargs.get('Qualifier', '')

        if extraneous_qualifier and arn_qualifier and arn_qualifier != extraneous_qualifier:
            raise ValueError('The derived qualifier from the function name does not match the specified qualifier.')

        final_qualifier = arn_qualifier if arn_qualifier else extraneous_qualifier

        try:
            # GGC v1.9.0 or newer
            function_arn = FunctionArnFields.build_function_arn(arn_fields.unqualified_arn, final_qualifier)
        except AttributeError:
            # older GGC version
            raise AttributeError('class FunctionArnFields has no attribute \'build_function_arn\'. build_function_arn '
                                 'is introduced in GGC v1.9.0. Please check your GGC version.')

        # ClientContext must be base64 if given, but is an option parameter
        try:
            client_context = kwargs.get('ClientContext', b'').decode()
        except AttributeError as e:
            customer_logger.exception(e)
            raise ValueError(
                '"ClientContext" argument must be a byte string or support a decode method which returns a string'
            )

        if client_context:
            if not re.match(valid_base64_regex, client_context):
                raise ValueError('"ClientContext" argument of Lambda.Client.invoke must be base64 encoded.')

        # Payload is an optional parameter
        payload = kwargs.get('Payload', b'')
        invocation_type = kwargs.get('InvocationType', 'RequestResponse')
        customer_logger.debug('Invoking local lambda "{}" with payload "{}" and client context "{}"'.format(
            function_arn, payload, client_context))

        # Post the work to IPC and return the result of that work
        return self._invoke_internal(function_arn, payload, client_context, invocation_type)

    @mock
    def _invoke_internal(self, function_arn, payload, client_context, invocation_type="RequestResponse"):
        """
        This private method is seperate from the main, public invoke method so that other code within this SDK can
        give this Lambda client a raw payload/client context to invoke with, rather than having it built for them.
        This lets you include custom ExtensionMap_ values like subject which are needed for our internal pinned Lambdas.
        """
        customer_logger.debug('Invoking Lambda function "{}" with Greengrass Message "{}"'.format(function_arn, payload))

        try:
            invocation_id = self.ipc.post_work(function_arn, payload, client_context, invocation_type)

            if invocation_type == "Event":
                # TODO: Properly return errors based on BOTO response
                # https://boto3.readthedocs.io/en/latest/reference/services/lambda.html#Lambda.Client.invoke
                return {'Payload': b'', 'FunctionError': ''}

            work_result_output = self.ipc.get_work_result(function_arn, invocation_id)
            if not work_result_output.func_err:
                output_payload = StreamingBody(work_result_output.payload)
            else:
                output_payload = work_result_output.payload
            invoke_output = {
                'Payload': output_payload,
                'FunctionError': work_result_output.func_err,
            }
            return invoke_output
        except IPCException as e:
            customer_logger.exception(e)
            raise InvocationException('Failed to invoke function due to ' + str(e))


class StreamingBody(object):
    """Wrapper class for http response payload

    This provides a consistent interface to AWS Lambda Python SDK
    """
    def __init__(self, payload):
        self._raw_stream = BytesIO(payload)
        self._amount_read = 0

    def read(self, amt=None):
        """Read at most amt bytes from the stream.
        If the amt argument is omitted, read all data.
        """
        chunk = self._raw_stream.read(amt)
        self._amount_read += len(chunk)
        return chunk

    def close(self):
        self._raw_stream.close()



#
# Copyright 2010-2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#

import json
from functools import wraps
from greengrass_common.env_vars import MY_FUNCTION_ARN


def mock(func):
    """
    mock decorates _invoke_internal by checking if MY_FUNCTION_ARN is present
    if MY_FUNCTION_ARN is present, the actual _invoke_internal is invoked
    otherwise, the mock _invoke_internal is invoked
    """
    @wraps(func)
    def mock_invoke_internal(self, function_arn, payload, client_context, invocation_type="RequestResponse"):
        if MY_FUNCTION_ARN is None:
            if invocation_type == 'RequestResponse':
                return {
                    'Payload': json.dumps({
                        'TestKey': 'TestValue'
                    }),
                    'FunctionError': ''
                }
            elif invocation_type == 'Event':
                return {
                    'Payload': b'',
                    'FunctionError': ''
                }
            else:
                raise Exception('Unsupported invocation type {}'.format(invocation_type))
        else:
            return func(self, function_arn, payload, client_context, invocation_type)
    return mock_invoke_internal



#
# Copyright 2010-2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#

import base64
import json
import logging

from greengrasssdk import Lambda
from greengrass_common.env_vars import SHADOW_FUNCTION_ARN, ROUTER_FUNCTION_ARN, MY_FUNCTION_ARN

# Log messages in the SDK are part of customer's log because they're helpful for debugging
# customer's lambdas. Since we configured the root logger to log to customer's log and set the
# propagate flag of this logger to True. The log messages submitted from this logger will be
# sent to the customer's local Cloudwatch handler.
customer_logger = logging.getLogger(__name__)
customer_logger.propagate = True


class ShadowError(Exception):
    pass


class Client:
    def __init__(self):
        self.lambda_client = Lambda.Client()

    def get_thing_shadow(self, **kwargs):
        r"""
        Call shadow lambda to obtain current shadow state.

        :Keyword Arguments:
            * *thingName* (``string``) --
              [REQUIRED]
              The name of the thing.

        :returns: (``dict``) --
        The output from the GetThingShadow operation
            * *payload* (``bytes``) --
              The state information, in JSON format.
        """
        thing_name = self._get_required_parameter('thingName', **kwargs)
        payload = b''

        return self._shadow_op('get', thing_name, payload)

    def update_thing_shadow(self, **kwargs):
        r"""
        Updates the thing shadow for the specified thing.

        :Keyword Arguments:
            * *thingName* (``string``) --
              [REQUIRED]
              The name of the thing.
            * *payload* (``bytes or seekable file-like object``) --
              [REQUIRED]
              The state information, in JSON format.

        :returns: (``dict``) --
        The output from the UpdateThingShadow operation
            * *payload* (``bytes``) --
              The state information, in JSON format.
        """
        thing_name = self._get_required_parameter('thingName', **kwargs)
        payload = self._get_required_parameter('payload', **kwargs)

        return self._shadow_op('update', thing_name, payload)

    def delete_thing_shadow(self, **kwargs):
        r"""
        Deletes the thing shadow for the specified thing.

        :Keyword Arguments:
            * *thingName* (``string``) --
              [REQUIRED]
              The name of the thing.

        :returns: (``dict``) --
        The output from the DeleteThingShadow operation
            * *payload* (``bytes``) --
              The state information, in JSON format.
        """
        thing_name = self._get_required_parameter('thingName', **kwargs)
        payload = b''

        return self._shadow_op('delete', thing_name, payload)

    def publish(self, **kwargs):
        r"""
        Publishes state information.

        :Keyword Arguments:
            * *topic* (``string``) --
              [REQUIRED]
              The name of the MQTT topic.
            * *payload* (``bytes or seekable file-like object``) --
              The state information, in JSON format.
            * *queueFullPolicy* (``string``) --
              The policy for GGC to take when its internal queue is full
        :returns: None
        """

        topic = self._get_required_parameter('topic', **kwargs)

        # payload and queueFullPolicy are optional parameters
        payload = kwargs.get('payload', b'')
        queue_full_policy = kwargs.get('queueFullPolicy', '')

        function_arn = ROUTER_FUNCTION_ARN
        client_context = {
            'custom': {
                'source': MY_FUNCTION_ARN,
                'subject': topic
            }
        }

        if queue_full_policy == 'AllOrException':
            client_context['custom']['queueFullPolicy'] = 'AllOrError'
        elif queue_full_policy in ['BestEffort', '']:
            client_context['custom']['queueFullPolicy'] = queue_full_policy
        else:
            raise ValueError('Invalid value for queueFullPolicy: {queueFullPolicy}'.format(
                queueFullPolicy=queue_full_policy
            ))

        customer_logger.debug('Publishing message on topic "{}" with Payload "{}"'.format(topic, payload))
        self.lambda_client._invoke_internal(
            function_arn,
            payload,
            base64.b64encode(json.dumps(client_context).encode()),
            'Event'
        )

    def _get_required_parameter(self, parameter_name, **kwargs):
        if parameter_name not in kwargs:
            raise ValueError('Parameter "{parameter_name}" is a required parameter but was not provided.'.format(
                parameter_name=parameter_name
            ))
        return kwargs[parameter_name]

    def _shadow_op(self, op, thing_name, payload):
        topic = '$aws/things/{thing_name}/shadow/{op}'.format(thing_name=thing_name, op=op)
        function_arn = SHADOW_FUNCTION_ARN
        client_context = {
            'custom': {
                'subject': topic
            }
        }

        customer_logger.debug('Calling shadow service on topic "{}" with payload "{}"'.format(topic, payload))
        response = self.lambda_client._invoke_internal(
            function_arn,
            payload,
            base64.b64encode(json.dumps(client_context).encode())
        )

        if response['FunctionError'] != '':
            raise ShadowError('Request for shadow state returned FunctionError "{}" with error payload "{}"'.format(
                response['FunctionError'], response['Payload']
            ))

        payload = response['Payload'].read()
        if response:
            response_payload_map = json.loads(payload.decode('utf-8'))
            if 'code' in response_payload_map and 'message' in response_payload_map:
                raise ShadowError('Request for shadow state returned error code {} with message "{}"'.format(
                    response_payload_map['code'], response_payload_map['message']
                ))

        return {'payload': payload}
